import os
import platform
import threading

from .menu_formatter import MenuFormatBuilder
from .screen import Screen


class ConsoleMenu(object):
    """
    A class that displays a menu and allows the user to select an option.

    Args:
        title (str): The title of the menu, or a method reference that returns a string.
        subtitle (str): The subtitle of the menu, or a method reference that returns a string.
        screen (:obj:`consolemenu.screen.Screen`): The screen object associated with this menu.
        formatter (:obj:`MenuFormatBuilder`): The MenuFormatBuilder instance used to format this menu.
        prologue_text (str): Text or method reference to include in the "prologue" section of the menu.
        epilogue_text (str): Text or method reference to include in the "epilogue" section of the menu.
        show_exit_option (bool): Specifies whether this menu should show an exit item by default. Defaults to True.
            Can be overridden when the menu is started.
        exit_option_text (str): Text for the Exit menu item. Defaults to 'Exit'.

    Attributes:
        cls.currently_active_menu (:obj:`ConsoleMenu`): Class variable that holds the currently active menu or None
            if no menu is currently active (e.g. when switching between menus)
        items (:obj:`list` of :obj:`MenuItem`): The list of MenuItems that the menu will display
        parent (:obj:`ConsoleMenu`): The parent of this menu
        previous_active_menu (:obj:`ConsoleMenu`): the previously active menu to be restored into the class's
            currently active menu
        current_option (int): The currently highlighted menu option
        selected_option (int): The option that the user has most recently selected
    """

    currently_active_menu = None

    def __init__(self, title=None, subtitle=None, screen=None, formatter=None,
                 prologue_text=None, epilogue_text=None,
                 show_exit_option=True, exit_option_text='Exit'):
        if screen is None:
            screen = Screen()
        self.screen = screen

        if formatter is None:
            formatter = MenuFormatBuilder()
        self.formatter = formatter

        self.title = title
        self.subtitle = subtitle
        self.prologue_text = prologue_text
        self.epilogue_text = epilogue_text

        self.highlight = None
        self.normal = None

        self.show_exit_option = show_exit_option

        self.items = list()

        self.parent = None

        self.exit_item = ExitItem(menu=self, text=exit_option_text)

        self.current_option = 0
        self.selected_option = -1

        self.returned_value = None

        self.should_exit = False

        self.previous_active_menu = None

        self._main_thread = None

        self._running = threading.Event()

    def __repr__(self):
        return "%s: %s. %d items" % (self.get_title(), self.get_subtitle(), len(self.items))

    @property
    def current_item(self):
        """
        :obj:`consolemenu.items.MenuItem`: The item corresponding to the menu option that is currently highlighted,
            or None.
        """
        if self.items:
            return self.items[self.current_option]
        else:
            return None

    @property
    def selected_item(self):
        """
        :obj:`consolemenu.items.MenuItem`:  The item in :attr:`items` that the user most recently selected, or None.
        """
        if self.items and self.selected_option != -1:
            return self.items[self.current_option]
        else:
            return None

    def append_item(self, item):
        """
        Add an item to the end of the menu before the exit item.

        Args:
            item (MenuItem): The item to be added.

        """
        did_remove = self.remove_exit()
        item.menu = self
        self.items.append(item)
        if did_remove:
            self.add_exit()

    def remove_item(self, item):
        """
        Remove the specified item from the menu.

        Args:
            item (MenuItem): the item to be removed.

        Returns:
            bool: True if the item was removed; False otherwise.
        """
        for idx, _item in enumerate(self.items):
            if item == _item:
                del self.items[idx]
                return True
        return False

    def add_exit(self):
        """
        Add the exit item if necessary. Used to make sure there aren't multiple exit items.

        Returns:
            bool: True if item needed to be added, False otherwise.
        """
        if not self.items or self.items[-1] is not self.exit_item:
            self.items.append(self.exit_item)
            return True
        return False

    def remove_exit(self):
        """
        Remove the exit item if necessary. Used to make sure we only remove the exit item, not something else.

        Returns:
            bool: True if item needed to be removed, False otherwise.
        """
        if self.items:
            if self.items[-1] is self.exit_item:
                del self.items[-1]
                return True
        return False

    def is_selected_item_exit(self):
        """
        Checks to determine if the currently selected item is the Exit Menu item.

        Returns:
            bool: True if the currently selected item is the Exit Menu item; False otherwise.
        """
        return self.selected_item and self.selected_item is self.exit_item

    def _wrap_start(self):
        self._main_loop()
        ConsoleMenu.currently_active_menu = None
        self.clear_screen()
        ConsoleMenu.currently_active_menu = self.previous_active_menu

    def start(self, show_exit_option=None):
        """
        Start the menu in a new thread and allow the user to interact with it.
        The thread is a daemon, so :meth:`join()<consolemenu.ConsoleMenu.join>` should be called if there's a
        possibility that the main thread will exit before the menu is done

        Args:
            show_exit_option (bool): Specify whether the exit item should be shown, defaults to the value
                set in the constructor

        """
        self.previous_active_menu = ConsoleMenu.currently_active_menu
        ConsoleMenu.currently_active_menu = None

        self.should_exit = False

        if show_exit_option is None:
            show_exit_option = self.show_exit_option

        if show_exit_option:
            self.add_exit()
        else:
            self.remove_exit()

        try:
            self._main_thread = threading.Thread(target=self._wrap_start, daemon=True)
        except TypeError:
            self._main_thread = threading.Thread(target=self._wrap_start)
            self._main_thread.daemon = True

        self._main_thread.start()

    def show(self, show_exit_option=None):
        """
        Calls start and then immediately joins.

        Args:
            show_exit_option (bool):  Specify whether the exit item should be shown, defaults to the value set
                in the constructor

        """
        self.start(show_exit_option)
        self.join()

    def _main_loop(self):
        self._set_up_colors()
        ConsoleMenu.currently_active_menu = self
        self._running.set()

        while self._running.wait() is not False and not self.should_exit:
            self.screen.clear()
            self.draw()
            self.process_user_input()

    def draw(self):
        """
        Refresh the screen and redraw the menu. Should be called whenever something changes that needs to be redrawn.
        """
        self.screen.printf(self.formatter.format(title=self.get_title(),
                                                 subtitle=self.get_subtitle(),
                                                 items=self.items,
                                                 prologue_text=self.get_prologue_text(),
                                                 epilogue_text=self.get_epilogue_text()))

    def is_running(self):
        """
        Check if the menu has been started and is not paused.

        Returns:
            bool: True if the menu is started and hasn't been paused; False otherwise.
        """
        return self._running.is_set()

    def wait_for_start(self, timeout=None):
        """
        Block until the menu is started.

        Args:
            timeout:  How long to wait before timing out.

        Returns:
            bool: False if timeout is given and operation times out, True otherwise. None before Python 2.7.
        """
        return self._running.wait(timeout)

    def is_alive(self):
        """
        Check whether the thread is stil alive.

        Returns:
            bool: True if the thread is still alive; False otherwise.
        """
        return self._main_thread.is_alive()

    def pause(self):
        """
        Temporarily pause the menu until resume is called.
        """
        self._running.clear()

    def resume(self):
        """
        Sets the currently active menu to this one and resumes it.
        """
        ConsoleMenu.currently_active_menu = self
        self._running.set()

    def join(self, timeout=None):
        """
        Should be called at some point after :meth:`start()<consolemenu.ConsoleMenu.start>` to block until
        the menu exits.

        Args:
            timeout (Number): How long to wait before timing out.

        """
        self._main_thread.join(timeout=timeout)

    def get_input(self):
        """
        Can be overridden to change the input method.
        Called in :meth:`process_user_input()<consolemenu.ConsoleMenu.process_user_input>`

        :return: the ordinal value of a single character
        :rtype: int
        """
        return self.screen.input()

    def process_user_input(self):
        """
        Gets the next single character and decides what to do with it
        """
        user_input = self.get_input()

        try:
            num = int(user_input)
        except Exception:
            return
        if 0 < num < len(self.items) + 1:
            self.current_option = num - 1
            self.select()

        return user_input

    def go_to(self, option):
        """
        Go to the option entered by the user as a number

        :param option: the option to go to
        :type option: int
        """
        self.current_option = option
        self.draw()

    def go_down(self):
        """
        Go down one, wrap to beginning if necessary
        """
        if self.current_option < len(self.items) - 1:
            self.current_option += 1
        else:
            self.current_option = 0
        self.draw()

    def go_up(self):
        """
        Go up one, wrap to end if necessary
        """
        if self.current_option > 0:
            self.current_option += -1
        else:
            self.current_option = len(self.items) - 1
        self.draw()

    def select(self):
        """
        Select the current item and run it
        """
        self.selected_option = self.current_option
        self.selected_item.set_up()
        self.selected_item.action()
        self.selected_item.clean_up()
        self.returned_value = self.selected_item.get_return()
        self.should_exit = self.selected_item.should_exit

    def exit(self):
        """
        Signal the menu to exit, then block until it's done cleaning up
        """
        self.should_exit = True
        self.join()

    def _set_up_colors(self):
        # TODO add color support
        # curses.init_pair(1, curses.COLOR_BLACK, curses.COLOR_WHITE)
        # self.highlight = curses.color_pair(1)
        # self.normal = curses.A_NORMAL
        pass

    def clear_screen(self):
        """
        Clear the screen belonging to this menu
        """
        self.screen.clear()

    # Getters to get text in case method reference
    def get_title(self):
        return self.title() if callable(self.title) else self.title

    def get_subtitle(self):
        return self.subtitle() if callable(self.subtitle) else self.subtitle

    def get_prologue_text(self):
        return self.prologue_text() if callable(self.prologue_text) else self.prologue_text

    def get_epilogue_text(self):
        return self.epilogue_text() if callable(self.epilogue_text) else self.epilogue_text


class MenuItem(object):
    """
    A generic menu item
    """

    def __init__(self, text, menu=None, should_exit=False):
        """
        :ivar str text: The text shown for this menu item
        :ivar ConsoleMenu menu: The menu to which this item belongs
        :ivar bool should_exit: Whether the menu should exit once this item's action is done
        """
        self.text = text
        self.menu = menu
        self.should_exit = should_exit

    def __str__(self):
        return "%s %s" % (self.menu.get_title(), self.get_text())

    def show(self, index):
        """
        How this item should be displayed in the menu. Can be overridden, but should keep the same signature.

        Default is:

            1 - Item 1

            2 - Another Item

        :param int index: The index of the item in the items list of the menu
        :return: The representation of the item to be shown in a menu
        :rtype: str
        """
        return "%2d - %s" % (index + 1, self.get_text())

    def set_up(self):
        """
        Override to add any setup actions necessary for the item
        """
        pass

    def action(self):
        """
        Override to carry out the main action for this item.
        """
        pass

    def clean_up(self):
        """
        Override to add any cleanup actions necessary for the item
        """
        pass

    def get_return(self):
        """
        Override to change what the item returns.
        Otherwise just returns the same value the last selected item did.
        """
        return self.menu.returned_value

    def __eq__(self, o):
        return self.text == o.text and self.menu == o.menu and self.should_exit == o.should_exit

    # Getters to get text in case method reference
    def get_text(self):
        return self.text() if callable(self.text) else self.text


class ExitItem(MenuItem):
    """
    Used to exit the current menu. Handled by :class:`consolemenu.ConsoleMenu`
    """

    def __init__(self, text="Exit", menu=None):
        super(ExitItem, self).__init__(text=text, menu=menu, should_exit=True)

    def show(self, index):
        """
        ExitItem overrides this method to display appropriate Exit or Return text.
        """
        # If we have a parent menu, and no overriding exit text was specified,
        # change Exit text to "Return to {Parent Menu Title}"
        if self.menu and self.menu.parent and self.get_text() == 'Exit':
            self.text = "Return to %s" % self.menu.parent.get_title()
        return super(ExitItem, self).show(index)


def clear_terminal():
    """
    Call the platform specific function to clear the terminal: cls on windows, reset otherwise
    """
    if platform.system().lower() == "windows":
        os.system('cls')
    else:
        os.system('reset')



import logging
import mmap
import os
import re
from typing import Dict, List, Optional

from ...model.ExternalLink import ExternalLink


class ExtParser(object):
    """Class responsible for parsing lab.ext file."""

    @staticmethod
    def parse(path: str) -> Optional[Dict[str, List[ExternalLink]]]:
        """Parse the lab.ext and return a Dict. Keys are name of collision domain and values are List of ExternalLink
        attached to that interface.

        Args:
            path (str): The path to lab.ext file.

        Returns:
            Optional[Dict[str, List[ExternalLink]]]: Keys are name of collision domain and values are List of
                ExternalLink attached to that interface.

        Raises:
            IOError: If lab.ext file cannot be opened.
            ValueError: If the VLAN ID is not allowed.
            SyntaxError: If lab.ext file is malformed.
        """
        lab_ext_path = os.path.join(path, 'lab.ext')

        if not os.path.exists(lab_ext_path):
            return None

        if os.stat(lab_ext_path).st_size == 0:
            logging.warning("lab.ext file is empty. Ignoring...")
            return None

        # Reads lab.ext in memory so it is faster.
        try:
            with open(lab_ext_path, 'r') as ext_file:
                ext_mem_file = mmap.mmap(ext_file.fileno(), 0, access=mmap.ACCESS_READ)
        except Exception:
            raise IOError("Cannot open lab.ext file.")

        external_links = {}

        line_number = 1
        line = ext_mem_file.readline().decode('utf-8')
        while line:
            # E.g. A enp9s0
            # B enp9s0.20
            matches = re.search(r"^(?P<link>\w+)\s+(?P<interface>\w+)(?P<vlan>\.\d+)?$",
                                line.strip()
                                )

            if matches:
                link = matches.group("link").strip()
                interface = matches.group("interface").strip()
                vlan = int(matches.group("vlan").strip().replace(".", "")) if matches.group("vlan") else None

                if vlan:
                    if 0 <= vlan >= 4095:
                        raise ValueError(f"In file lab.ext, line {line_number}: VLAN ID must be in range [1, 4094].")

                if link not in external_links:
                    external_links[link] = []

                external_links[link].append(ExternalLink(interface, vlan))
            elif not line.startswith('#') and line.strip():
                raise SyntaxError(f"In file lab.ext - Line {line_number}.")

            line_number += 1
            line = ext_mem_file.readline().decode('utf-8')

        return external_links



from ..exceptions import DockerImageNotFoundError, InvalidImageArchitectureError
from ..setting.Setting import Setting
from ..trdparty.consolemenu.validators.base import BaseValidator


class ImageValidator(BaseValidator):
    def __init__(self) -> None:
        super(ImageValidator, self).__init__()

    def validate(self, input_string: str) -> bool:
        try:
            Setting.get_instance().check_image(input_string)
            return True
        except (ConnectionError, DockerImageNotFoundError, InvalidImageArchitectureError) as e:
            print(str(e))
            return False



from abc import ABC, abstractmethod
from typing import Dict, Any


class SettingsAddon(ABC):
    def load(self, settings: Dict[str, Any]) -> None:
        for name, value in settings.items():
            if hasattr(self, name):
                setattr(self, name, value)

    def get(self, name: str) -> Any:
        if hasattr(self, name):
            return getattr(self, name)
        else:
            raise AttributeError

    def merge(self, settings: Dict[str, Any] = None) -> Dict[str, Any]:
        to_dict = self._to_dict()

        if settings:
            settings.update(to_dict)
            return settings

        return to_dict

    @abstractmethod
    def _to_dict(self) -> Dict[str, Any]:
        pass



from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import Any, Generator


class IExecStream(ABC):
    """Interface for handling the commands stream exec

    Attributes:
        _stream (Generator): The generator yielding the output of the stream exec.
        _stream_api_object (Any): The specific API object to interact with the underlying stream.
    """
    __slots__ = ['_stream', '_stream_api_object']

    def __init__(self, stream: Generator, stream_api_object: Any) -> None:
        self._stream: Generator = stream
        self._stream_api_object: Any = stream_api_object

    @abstractmethod
    def stream_next(self) -> Iterator:
        """Return the next element from the stream.

        Returns:
            Iterator: The output iterator from the stream.
        """
        raise NotImplementedError("You must implement `next` method.")

    @abstractmethod
    def exit_code(self) -> int:
        """Return the exit code of the execution.

        Returns:
            int: The exit code of the execution.
        """
        raise NotImplementedError("You must implement `exit_code` method.")

    def __next__(self) -> Iterator:
        """Return the next element from the stream.
        This magic method allows to keep compatibility with the previous API.

        Returns:
            Iterator: The output iterator from the stream.
        """
        return self.stream_next()



from setuptools import find_packages, setup

setup(
    name='kathara',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    py_modules=['kathara'],
    version='3.8.1.post2',
    license='gpl-3.0',
    description='A lightweight container-based network emulation tool.',
    author='Kathara Framework',
    author_email='contact@kathara.org',
    url='https://www.kathara.org',
    download_url='https://github.com/KatharaFramework/Kathara/archive/refs/tags/3.8.1.tar.gz',
    keywords=['NETWORK-EMULATION', 'CONTAINERS', 'NFV'],
    install_requires=[
        "binaryornot>=0.4.4",
        "docker>=7.0.0",
        "kubernetes>=23.3.0",
        "requests>=2.22.0",
        "pyroute2",
        "rich",
        "setuptools<81",
        "fs>=2.4.16",
        "chardet<6.0.0",
        "libtmux>=0.8.2; platform_system == 'darwin' or platform_system == 'linux'",
        "appscript>=1.1.0; platform_system == 'darwin'",
        "pypiwin32>=223; platform_system == 'win32'",
        "windows-curses>=2.1.0; platform_system == 'win32'"
    ],
    extras_require={
        'test': ["pytest"],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
)



from . import items
from .console_menu import ConsoleMenu
from .console_menu import Screen
from .console_menu import clear_terminal
from .menu_formatter import MenuFormatBuilder
from .multiselect_menu import MultiSelectMenu
from .prompt_utils import PromptUtils
from .prompt_utils import UserQuit
from .selection_menu import SelectionMenu
from .version import __version__

__all__ = ['ConsoleMenu', 'SelectionMenu', 'MultiSelectMenu', 'MenuFormatBuilder', 'PromptUtils',
           'Screen', 'items', 'clear_terminal']



import asyncio
from typing import Optional

from .IConsoleAdapter import IConsoleAdapter
from .ITerminalSession import ITerminalSession
from .....utils import exec_by_platform


class TerminalRunner(object):
    """Bridge class between a console adapter and a terminal session.

    Attributes:
        console (IConsoleAdapter): OS-specific console adapter.
        session (ITerminalSession): Wrapper of the manager terminal session.
        _loop (Optional[asyncio.AbstractEventLoop]): Event loop from asyncio used by the runner.
        _closed (bool): Boolean flag indicating whether the runner is terminated.
        _tasks (list[asyncio.Task]): List of asyncio tasks created by the runner.
        _session_fd (Optional[int]): File descriptor from session.fileno() when available.
    """
    __slots__ = ["console", "session", "_loop", "_closed", "_tasks", "_session_fd"]

    def __init__(self, console: IConsoleAdapter, session: ITerminalSession) -> None:
        self.console: IConsoleAdapter = console
        self.session: ITerminalSession = session

        self._loop: Optional[asyncio.AbstractEventLoop] = None

        self._closed: bool = False

        self._tasks: list[asyncio.Task] = []
        self._session_fd: Optional[int] = None

    def start(self) -> None:
        """Start the bridge and block until the runner stops.

        This method:
            - creates and sets a new asyncio loop
            - enters console raw mode
            - starts watching resize and input
            - starts pumping session output to stdout
            - runs the loop forever until close() is invoked

        Returns:
            None
        """
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)

        self.console.enter_raw()
        self.console.watch_resize(self._loop, self._on_resize)

        self.console.install_input_reader(self._loop, self._on_stdin_bytes, self.close)
        self._start_stdout()

        try:
            self._loop.run_forever()
        finally:
            self._loop.close()

    def _on_resize(self, cols: int, rows: int) -> None:
        """Handle terminal resize events.

        Args:
            cols (int): Terminal width in columns.
            rows (int): Terminal height in rows.

        Returns:
            None
        """
        if self._closed:
            return

        try:
            self.session.resize(cols, rows)
        except Exception:
            self.close()

    def _on_stdin_bytes(self, data: bytes) -> None:
        """Handle bytes received from local stdin.

        Args:
            data (bytes): Data read from stdin.

        Returns:
            None
        """
        if self._closed:
            return

        if not data:
            self.close()
            return

        try:
            self.session.write(data)
        except Exception:
            self.close()

    def _on_session_readable(self) -> None:
        """Handle readiness notification for session output (fd-based). Called when the session fd becomes readable.

        Returns:
            None
        """
        if self._closed:
            return

        try:
            data = self.session.read(4096)
        except Exception:
            self.close()
            return

        if not data:
            self.close()
            return

        try:
            self.console.write_stdout(data)
        except Exception:
            self.close()
            return

    async def _threaded_stdout(self) -> None:
        """Continuously read session output in a thread executor and write to stdout.

        This is used when:
            - the session does not expose a file descriptor (fileno() is None), or
            - on Windows where fd-based readers are not unavailable.

        Returns:
            None
        """
        while not self._closed:
            try:
                data = await self._loop.run_in_executor(None, self.session.read, 4096)
            except Exception:
                self.close()
                return

            if self._closed:
                return

            if not data:
                await asyncio.sleep(0.03)
                continue

            try:
                self.console.write_stdout(data)
            except Exception:
                self.close()
                return

    def _start_stdout(self) -> None:
        """Start pumping session output to local stdout using an OS-specify strategy.

        Returns:
            None
        """

        def unix():
            self._session_fd = self.session.fileno()
            if self._session_fd is None:
                self._tasks.append(self._loop.create_task(self._threaded_stdout()))
                return

            self._loop.add_reader(self._session_fd, self._on_session_readable)

        def windows():
            self._tasks.append(self._loop.create_task(self._threaded_stdout()))

        exec_by_platform(unix, windows, unix)

    def close(self) -> None:
        """Stop the runner and perform cleanup.

        Returns:
            None
        """
        if self._closed or self._loop is None:
            return

        self._closed = True
        self._cleanup()
        self._loop.stop()

    def _cleanup(self) -> None:
        """Internal cleanup routine.

        Attempts best-effort cleanup of:
            - all asyncio tasks
            - session fd readers
            - resize watchers and stdin readers
            - session close
            - console raw mode restoration

        Returns:
            None
        """
        if self._loop is None:
            return

        for task in asyncio.all_tasks(self._loop):
            try:
                task._log_destroy_pending = False
            except Exception:
                pass
            task.cancel()

        for task in self._tasks:
            try:
                task._log_destroy_pending = False
            except Exception:
                pass
            task.cancel()

        if self._session_fd is not None:
            try:
                self._loop.remove_reader(self._session_fd)
            except Exception:
                pass
            self._session_fd = None

        try:
            self.console.unwatch_resize(self._loop)
        except Exception:
            pass

        try:
            self.console.remove_input_reader(self._loop)
        except Exception:
            pass

        try:
            self.session.close()
        except Exception:
            pass

        try:
            self.console.exit_raw()
        except Exception:
            pass



from enum import IntEnum


class SharedCollisionDomainsOption(IntEnum):
    """Enum representing options for shared collision domains option.

    Attributes:
        NOT_SHARED (int): Represents the option for not sharing collision domains (value: 1).
        LABS (int): Represents the option for sharing collision domains among network scenarios of the same user
            (value: 2).
        USERS (int): Represents the option for sharing collision domains among network scenarios of different users
            (value: 3).
    """
    NOT_SHARED = 1
    LABS = 2
    USERS = 3

    @staticmethod
    def to_string(value):
        if value == 1:
            return "Not Shared"
        elif value == 2:
            return "Share collision domains between network scenarios"
        elif value == 3:
            return "Share collision domains between users"



from .menu_borders import AsciiBorderStyle
from .menu_borders import DoubleLineBorderStyle
from .menu_borders import DoubleLineOuterLightInnerBorderStyle
from .menu_borders import HeavyBorderStyle
from .menu_borders import HeavyOuterLightInnerBorderStyle
from .menu_borders import LightBorderStyle
from .menu_borders import MenuBorderStyle
from .menu_borders import MenuBorderStyleFactory
from .menu_borders import MenuBorderStyleType
from .menu_margins import MenuMargins
from .menu_padding import MenuPadding
from .menu_style import MenuStyle

__all__ = ['MenuBorderStyle', 'MenuBorderStyleType', 'MenuBorderStyleFactory', 'MenuMargins', 'MenuPadding',
           'MenuStyle', 'AsciiBorderStyle', 'LightBorderStyle', 'HeavyBorderStyle', 'DoubleLineBorderStyle',
           'DoubleLineOuterLightInnerBorderStyle', 'HeavyOuterLightInnerBorderStyle']



import logging

from rich.console import Console
from rich.tree import Tree

from ...ui.utils import confirmation_prompt
from ....model.Lab import Lab
from ....model.Machine import Machine
from ....setting.Setting import Setting


class MountDevicesVolumes(object):
    """Listener fired when devices have volumes to mount."""

    def run(self, lab: Lab, machines_with_volumes: dict[str, Machine]) -> None:
        """Prompt the user for confirming devices volumes.

        Args:
            lab (Lab): The network scenario.
            machines_with_volumes (dict[str, Machine]): Dict containing Machine objects with volumes attached.

        Returns:
            None
        """
        console = Console()
        console.print("The following devices have volumes configured:")
        for machine_name, machine in machines_with_volumes.items():
            tree = Tree(f"* [cyan]Device `{machine_name}`[/cyan]")
            for host_path, volume in machine.meta["volumes"].items():
                guest_path = volume.get("guest_path", "<missing guest_path>")
                tree.add(f"Host Path: [green]{host_path}[/green] -> Device Path: [magenta]{guest_path}[/magenta]")
            console.print(tree)

        policy = Setting.get_instance().volume_mount_policy
        if policy == 'Prompt':
            confirmation_prompt(
                "Continue with volume mounting?",
                lambda: None,
                lambda: lab.add_option("_mount_volumes", False)
            )



import json
from typing import Dict, Any

from ....foundation.manager.stats.ILinkStats import ILinkStats


class KubernetesLinkStats(ILinkStats):
    """The class responsible to handle Kubernetes Networks statistics.

    Attributes:
        link_api_object (Any): The Kubernetes Network associated with this statistics.
        lab_hash (str): The hash identifier of the network scenario of the Kubernetes Network.
        name (str): The name of the collision domain.
        network_name (str): The Kubernetes Network Name.
        vxlan_id (str): The VXLAN Identifier associated to this network.
    """
    __slots__ = ['link_api_object', 'lab_hash', 'name', 'network_name', 'vxlan_id']

    def __init__(self, link_api_object: Any):
        self.link_api_object: Any = link_api_object
        self.lab_hash: str = link_api_object['metadata']['namespace']
        self.name: str = link_api_object['metadata']['labels']['name']
        self.network_name: str = link_api_object['metadata']['name']
        self.vxlan_id = json.loads(link_api_object['spec']['config'])['vxlanId']

    def update(self) -> None:
        """Update dynamic statistics with the current ones.

        Returns:
            None
        """
        return

    def to_dict(self) -> Dict[str, Any]:
        """Transform statistics into a dict representation.

        Returns:
            Dict[str, Any]: Dict containing statistics.
        """
        return {
            "network_scenario_id": self.lab_hash,
            "name": self.name,
            "network_name": self.network_name,
            "vxlan_id": self.vxlan_id
        }

    def __repr__(self) -> str:
        return str(self.to_dict())

    def __str__(self) -> str:
        """Return a formatted string with the link statistics.

        Returns:
           str: A formatted string with the link statistics
        """
        formatted_stats = f"Network Scenario ID: {self.lab_hash}"
        formatted_stats += f"\nLink Name: {self.name}"
        formatted_stats += f"\nNetwork Name: {self.network_name}"
        formatted_stats += f"\nVXLAN ID: {self.vxlan_id}"

        return formatted_stats



import io
import os.path
import re
from typing import Optional, List, BinaryIO, TextIO, Union

from fs import open_fs
from fs.base import FS
from fs.copy import copy_dir

from ...exceptions import InvocationError


class FilesystemMixin(object):
    """Abstraction to manage filesystems of devices and network scenarios.

    Attributes:
        fs (FS): An object referencing a filesystem. Can be both real OS or a memory fs.
    """

    __slots__ = ['fs']

    def __init__(self):
        self.fs: Optional[FS] = None

    def fs_type(self) -> Optional[str]:
        """Return the name of the class of the fs object, if present. Else, return None.

        Returns:
            Optional[str]: The name of the class of the fs object.
        """
        return self.fs.__class__.__name__.lower().replace("fs", "") if self.fs else None

    def fs_path(self) -> Optional[str]:
        """Return the path of the filesystem, if fs has a path on the host. Else, return None

        Returns:
            Optional[str]: The path of the filesystem in the fs.
        """
        return (self.fs.getsyspath("") if self.fs.hassyspath("") else None) if self.fs else None

    def create_file_from_string(self, content: str, dst_path: str) -> None:
        """Create a file in the fs object from a string.

        Args:
            content (str): The string representing the content of the file to create.
            dst_path (str): The absolute path of the fs where create the file.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
            fs.errors.ResourceNotFound: If the path is not found in the fs.
        """
        if not self.fs:
            raise InvocationError("Cannot create a file if the filesystem is not set.")

        directory = os.path.dirname(dst_path)
        self.fs.makedirs(directory, recreate=True)

        with self.fs.open(dst_path, "w") as dst_file:
            dst_file.write(content)

    def update_file_from_string(self, content: str, dst_path: str) -> None:
        """Update a file in the fs object from a string.

        Args:
            content (str): The string representing the content for updating the file.
            dst_path (str): The absolute path on the fs of the file to update.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
            fs.errors.ResourceNotFound: If the path is not found in the fs.
        """
        if not self.fs:
            raise InvocationError("There is no filesystem associated to this object.")

        with self.fs.open(dst_path, "a") as dst_file:
            dst_file.write(content)

    def create_file_from_list(self, lines: List[str], dst_path: str) -> None:
        """Create a file in the fs object from a list of strings.

        Args:
            lines (List[str]): The list of strings representing the content of the file to create.
            dst_path (str): The absolute path of the fs where create the file.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
            fs.errors.ResourceNotFound: If the path is not found in the fs.
        """
        if not self.fs:
            raise InvocationError("There is no filesystem associated to this object.")

        directory = os.path.dirname(dst_path)
        self.fs.makedirs(directory, recreate=True)

        with self.fs.open(dst_path, "w") as dst_file:
            dst_file.writelines(line + '\n' for line in lines)

    def update_file_from_list(self, lines: List[str], dst_path: str) -> None:
        """Update a file in the fs object from a list of strings.

        Args:
            lines (List[str]): The list of strings representing the content for updating the file.
            dst_path (str): The absolute path on the fs of the file to upload.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
            fs.errors.ResourceNotFound: If the path is not found in the fs.
        """
        if not self.fs:
            raise InvocationError("There is no filesystem associated to this object.")

        with self.fs.open(dst_path, "a") as dst_file:
            dst_file.writelines(line + '\n' for line in lines)

    def create_file_from_path(self, src_path: str, dst_path: str) -> None:
        """Create a file in the fs object from an existing file on the host filesystem.

        Args:
            src_path (str): The path of the file on the host filesystem to copy in the fs object.
            dst_path (str): The absolute path of the fs where create the file.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
            fs.errors.ResourceNotFound: If the path is not found in the fs.
        """
        if not self.fs:
            raise InvocationError("There is no filesystem associated to this object.")

        directory = os.path.dirname(dst_path)
        self.fs.makedirs(directory, recreate=True)

        with open(src_path, "rb") as dst_file:
            self.fs.upload(dst_path, dst_file)

    def create_file_from_stream(self, stream: Union[BinaryIO, TextIO], dst_path: str) -> None:
        """Create a file in the fs object from a stream.

        Args:
            stream (Union[BinaryIO, TextIO]): The stream representing the content of the file to create.
            dst_path (str): The absolute path of the fs where create the file.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
            UnsupportedOperation: If the stream is opened without read permissions.
            fs.errors.ResourceNotFound: If the path is not found in the fs.
        """
        if not self.fs:
            raise InvocationError("There is no filesystem associated to this object.")

        directory = os.path.dirname(dst_path)
        self.fs.makedirs(directory, recreate=True)

        try:
            if "b" in stream.mode:
                self.fs.upload(dst_path, stream)
            else:
                with self.fs.open(dst_path, "w") as dst_file:
                    dst_file.writelines(stream.readlines())
        except io.UnsupportedOperation:
            raise io.UnsupportedOperation("To create a file from stream, you must open it with read permissions.")

    def copy_directory_from_path(self, src_path: str, dst_path: str) -> None:
        """Copy a directory from a src_path in the host filesystem into a dst_path in this fs.

        Args:
             src_path (str): The source path of the directory to copy.
             dst_path (str): The destination path where to copy the directory.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
        """
        if not self.fs:
            raise InvocationError("There is no filesystem associated to this object.")

        directory_fs = open_fs(f"osfs://{os.path.abspath(src_path)}")
        copy_dir(directory_fs, ".", self.fs, dst_path)

    def write_line_before(self, file_path: str, line_to_add: str, searched_line: str,
                          first_occurrence: bool = False) -> int:
        """Write a new line before a specific line in a file.

        Args:
            file_path (str): The path of the file to add the new line.
            line_to_add (str): The new line to add before the searched line.
            searched_line (str): A string or a regex representing the searched line.
            first_occurrence (bool): Inserts line only before the first occurrence. Default is False.

        Returns:
            int: Number of times the line has been added.

        Raises:
            InvocationError: If the fs is None.
            fs.errors.FileExpected: If the path is not a file.
            fs.errors.ResourceNotFound: If the path does not exist.
        """
        if not self.fs:
            raise InvocationError("There is no filesystem associated to this object.")

        pattern = re.compile(searched_line)
        n_added = 0
        with self.fs.open(file_path, "r+") as file:
            file_lines = file.readlines()
            file.seek(0)
            file.truncate()
            new_lines = []
            for line in file_lines:
                if pattern.search(line):
                    if not first_occurrence or (first_occurrence and n_added == 0):
                        new_lines.append(line_to_add + '\n')
                        n_added += 1
                new_lines.append(line.replace("\n\r", "\n").replace("\r\n", "\n"))
            file.writelines(new_lines)

        return n_added

    def write_line_after(self, file_path: str, line_to_add: str, searched_line: str,
                         first_occurrence: bool = False) -> int:
        """Write a new line after a specific line in a file.

        Args:
            file_path (str): The path of the file to add the new line.
            line_to_add (str): The new line to add after the searched line.
            searched_line (str): A string or a regex representing the searched line.
            first_occurrence (bool): Inserts line only after the first occurrence. Default is False.

        Returns:
            int: Number of times the line has been added.

        Raises:
            InvocationError: If the fs is None.
            fs.errors.FileExpected: If the path is not a file.
            fs.errors.ResourceNotFound: If the path does not exist.
        """
        if not self.fs:
            raise InvocationError("There is no filesystem associated to this object.")

        pattern = re.compile(searched_line)
        n_added = 0
        with self.fs.open(file_path, "r+") as file:
            file_lines = file.readlines()
            file.seek(0)
            file.truncate()
            new_lines = []
            for line in file_lines:
                new_lines.append(line.replace("\n\r", "\n").replace("\r\n", "\n"))
                if pattern.search(line):
                    if not first_occurrence or (first_occurrence and n_added == 0):
                        new_lines.append(("\n" if not line.endswith("\n") else "") + line_to_add + '\n')
                        n_added += 1

            file.writelines(new_lines)

        return n_added

    def delete_line(self, file_path: str, line_to_delete: str, first_occurrence: bool = False) -> int:
        """Delete a specified line in a file.

        Args:
            file_path (str): The path of the file to delete the line.
            line_to_delete (str): A string or a regex representing the line to delete.
            first_occurrence (bool): Deletes only first occurrence. Default is False.

        Returns:
            int: Number of times the line has been deleted.

        Raises:
            InvocationError: If the fs is None.
            fs.errors.FileExpected: If the path is not a file.
            fs.errors.ResourceNotFound: If the path does not exist.
        """
        if not self.fs:
            raise InvocationError("There is no filesystem associated to this object.")

        pattern = re.compile(line_to_delete)
        n_deleted = 0
        with self.fs.open(file_path, "r+") as file:
            file_lines = file.readlines()
            file.seek(0)
            file.truncate()
            new_lines = []
            for line in file_lines:
                if pattern.search(line):
                    if not first_occurrence or (first_occurrence and n_deleted == 0):
                        n_deleted += 1
                        continue

                new_lines.append(line.replace("\n\r", "\n").replace("\r\n", "\n"))

            file.writelines(new_lines)

        return n_deleted
